import pprint
from lxml.etree import _Element


def _display_element(obj: _Element) -> str:
    elem_str = f"<{obj.tag} "

    # := if we drop 3.7
    id_str = obj.get("id")
    class_str = obj.get("class")

    if id_str:
        elem_str += f"id='{id_str}'"
    elif class_str:
        elem_str += f"class='{class_str}'"
    else:
        elem_str += " ".join(f"{k}='{v}'" for k, v in obj.attrib.items())

    return elem_str + f"> @ line {obj.sourceline}"


def _display(obj) -> str:
    if isinstance(obj, dict):
        return pprint.pformat(obj)
    elif hasattr(obj, "to_dict"):
        return pprint.pformat(obj.to_dict())
    elif isinstance(obj, _Element):
        return _display_element(obj)
    else:
        return repr(obj)



import dataclasses
import importlib
import typing
import lxml.html
import click
from scrapelib import Scraper
from .utils import _display
from .core import URL

try:
    from attr import has as attr_has
    from attr import fields as attr_fields
except ImportError:
    attr_has = lambda x: False  # noqa
    attr_fields = lambda x: []  # noqa


VERSION = "0.4.0"


def get_class(dotted_name: str):
    mod_name, cls_name = dotted_name.rsplit(".", 1)
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)


@click.group()
@click.version_option(version=VERSION)
def cli() -> None:
    pass


@cli.command()
@click.argument("url")
@click.option(
    "-ua",
    "--user-agent",
    default=f"spatula {VERSION}",
    help="override default user-agent",
)
@click.option("-X", "--verb", default="GET", help="set HTTP verb such as POST")
def shell(url: str, user_agent: str, verb: str) -> None:
    """
    Start an interactive Python session to interact with a particular page.
    """
    try:
        from IPython import embed
    except ImportError:
        print("shell command requires IPython")
        return

    # import selectors so they can be used without import
    from .selectors import SelectorError, XPath, SimilarLink, CSS  # noqa

    scraper = Scraper()
    scraper.user_agent = user_agent
    resp = scraper.request(verb, url)
    root = lxml.html.fromstring(resp.content)  # noqa
    click.secho(f"spatula {VERSION} shell", fg="blue")
    click.secho("available selectors: CSS, SimilarLink, XPath", fg="blue")
    click.secho("local variables", fg="green")
    click.secho("---------------", fg="green")
    click.secho("url: %s" % url, fg="green")
    click.secho("resp: requests Response instance", fg="green")
    click.secho(f"root: `lxml HTML element` <{root.tag}>", fg="green")
    embed()


def _get_fake_input(Cls, data, interactive):
    # build fake input from command line data if present
    fake_input = {}
    for item in data:
        k, v = item.split("=", 1)
        fake_input[k] = v

    input_type = getattr(Cls, "input_type", None)

    if hasattr(Cls, "example_input"):
        return getattr(Cls, "example_input")

    if input_type:
        print(f"{Cls.__name__} expects input ({input_type.__name__}): ")
        if dataclasses.is_dataclass(input_type):
            fields = dataclasses.fields(input_type)
        elif attr_has(input_type):
            fields = attr_fields(input_type)
        for field in fields:
            if field.name in fake_input:
                print(f"  {field.name}: {fake_input[field.name]}")
            elif interactive:
                fake_input[field.name] = click.prompt("  " + field.name)
            else:
                dummy_val = f"~{field.name}"
                fake_input[field.name] = dummy_val
                print(f"  {field.name}: {dummy_val}")
        return input_type(**fake_input)
    else:
        return fake_input


@cli.command()
@click.argument("class_name")
@click.option("-i", "--interactive", help="Interactively prompt for missing data.")
@click.option(
    "-d", "--data", multiple=True, help="Provide input data in name=value pairs."
)
@click.option("-s", "--source", help="Provide (or override) source URL")
def test(
    class_name: str, interactive: bool, data: typing.List[str], source: str
) -> None:
    """
    This command allows you to scrape a single page and see the output immediately.  This eases the common cycle of making modifications to a scraper, running a scrape (possibly with long-running but irrelevant portions commented out), and comparing output to what is expected.
    ``test`` can also be useful for debugging existing scrapers, you can see exactly what a single step of the scrape is providing, to help narrow down where erroneous data is coming from.

    Example::

    $ spatula test path.to.ClassName --source https://example.com

    This will run the scraper defined at :py:class:`path.to.ClassName` against the provided URL.
    """
    Cls = get_class(class_name)
    s = Scraper()

    fake_input = _get_fake_input(Cls, data, interactive)

    # special case for passing a single URL source
    if source:
        source = URL(source)
    if not source and hasattr(Cls, "example_source"):
        source = Cls.example_source

    # we need to do the request-response-next-page loop at least once
    once = True
    while source or once:
        once = False
        page = Cls(fake_input, source=source)

        # fetch data after input is handled, since we might need to build the source
        page._fetch_data(s)

        result = page.process_page()

        if isinstance(result, typing.Generator):
            for i, item in enumerate(result):
                print(f"{i}:", _display(item))
        else:
            print(_display(result))

        # will be None in most cases, existing the loop, otherwise we restart
        source = page.get_next_source()
        if source:
            click.secho(
                f"paginating for {page.__class__.__name__}: {source}", fg="blue"
            )


@cli.command()
@click.argument("workflow_name")
@click.option(
    "-o", "--output-dir", default=None, help="override default output directory."
)
def scrape(workflow_name: str, output_dir: str) -> None:
    """
    Run full workflow, and output data to disk.
    """
    workflow = get_class(workflow_name)
    workflow.execute(output_dir=output_dir)


if __name__ == "__main__":
    cli()



import io
import csv
import tempfile
import subprocess
import logging
import typing
import lxml.html
import scrapelib
from .core import URL, HandledError


def page_to_items(scraper, page):
    # fetch data for a page, and then call the process_page entrypoint
    page._fetch_data(scraper)
    result = page.process_page()

    # if we got back a generator, we need to process each result
    if isinstance(result, typing.Generator):
        # each item yielded might be a Page or an end-result
        for item in result:
            if isinstance(item, Page):
                yield from page_to_items(scraper, item)
            else:
                yield item

        # after handling a list, check for pagination
        next_source = page.get_next_source()
        if next_source:
            # instantiate the same class with same input, but increment the source
            yield from page_to_items(
                scraper, type(page)(page.input, source=next_source)
            )
    elif isinstance(result, Page):
        # single Page result, recurse deeper
        yield from page_to_items(scraper, result)
    else:
        # end-result, just return as-is
        yield result


class Page:
    source = None
    dependencies = {}

    def _fetch_data(self, scraper: scrapelib.Scraper) -> None:
        """
        ensure that the page has all of its data, this is guaranteed to be called exactly once
        before process_page is invoked
        """
        # process dependencies first
        for val, dep in self.dependencies.items():
            dep._fetch_data(scraper)
            setattr(self, val, dep.process_page())

        if not self.source:
            if hasattr(self, "get_source_from_input"):
                self.source = self.get_source_from_input()
            elif hasattr(self.input, "url"):
                self.source = URL(self.input.url)
            else:
                raise Exception(
                    f"{self.__class__.__name__} has no source or get_source_from_input"
                )
        if isinstance(self.source, str):
            self.source = URL(self.source)
        self.logger.info(f"fetching {self.source}")
        try:
            self.response = self.source.get_response(scraper)
        except scrapelib.HTTPError as e:
            self.handle_error_response(e)
            raise HandledError(e)
        else:
            self.postprocess_response()

    def __init__(self, input_val=None, *, source=None):
        self.input = input_val
        # allow possibility to override default source, useful during dev
        if source:
            self.source = source
        self.logger = logging.getLogger(self.__class__.__name__)

    def postprocess_response(self) -> None:
        """
        To be overridden.

        This is called after source.get_response but before self.process_page.
        """
        pass

    def process_error_response(self, exception) -> None:
        """
        To be overridden.

        This is called after source.get_response if an exception is raised.
        """
        pass

    def process_page(self):
        """
        To be overridden.

        Return data extracted from this page and this page alone.
        """
        raise NotImplementedError()

    def get_next_source(self):
        """
        To be overriden for paginated pages.

        Return a URL or valid source to fetch the next page, None if there isn't one.
        """
        return None


class HtmlPage(Page):
    """
    self.root: preprocessed lxml.html-parsed HTML element
    """

    def postprocess_response(self) -> None:
        self.root = lxml.html.fromstring(self.response.content)
        if hasattr(self.source, "url"):
            self.root.make_links_absolute(self.source.url)


class XmlPage(Page):
    """
    self.root: preprocessed lxml.etree-parsed XML element
    """

    def postprocess_response(self) -> None:
        self.root = lxml.etree.fromstring(self.response.content)


class JsonPage(Page):
    """
    self.data: preprocessed JSON
    """

    def postprocess_response(self) -> None:
        self.data = self.response.json()


class PdfPage(Page):
    preserve_layout = False

    def postprocess_response(self) -> None:
        with tempfile.NamedTemporaryFile() as temp:
            temp.write(self.response.content)
            temp.flush()
            temp.seek(0)

            if self.preserve_layout:
                command = ["pdftotext", "-layout", temp.name, "-"]
            else:
                command = ["pdftotext", temp.name, "-"]

            try:
                pipe = subprocess.Popen(
                    command, stdout=subprocess.PIPE, close_fds=True
                ).stdout
            except OSError as e:
                raise EnvironmentError(
                    f"error running pdftotext, missing executable? [{e}]"
                )
            data = pipe.read()
            pipe.close()
        self.text = data.decode("utf8")


class ListPage(Page):
    class SkipItem(Exception):
        pass

    def skip(self) -> None:
        raise self.SkipItem()

    def process_item(self, item):
        return item


class CsvListPage(ListPage):
    def postprocess_response(self) -> None:
        self.reader = csv.DictReader(io.StringIO(self.response.text))

    def process_page(self):
        for item in self.reader:
            try:
                item = self.process_item(item)
            except self.SkipItem:
                continue
            yield item

    def process_item(self, item):
        return item


class LxmlListPage(ListPage):
    """
    Base class for XML and HTML subclasses below, only difference is which parser is used.

    Simplification for pages that get a list of items and process them.

    When overriding the class, instead of providing process_page, one must only provide
    a selector and a process_item function.
    """

    selector = None

    def process_page(self):
        if not self.selector:
            raise NotImplementedError("must either provide selector or override scrape")
        items = self.selector.match(self.root)
        for item in items:
            try:
                item = self.process_item(item)
            except self.SkipItem:
                continue
            yield item


class HtmlListPage(LxmlListPage, HtmlPage):
    pass


class XmlListPage(LxmlListPage, XmlPage):
    pass


class JsonListPage(ListPage, JsonPage):
    def process_page(self):
        for item in self.data:
            try:
                item = self.process_item(item)
            except self.SkipItem:
                continue
            yield item



import os
import json
import glob
import datetime
import uuid
from typing import Optional
import requests
import scrapelib


class HandledError(Exception):
    def __init__(self, original):
        self.original = original


class Workflow:
    """
    Define a complete workflow by which items can be scraped and saved to disk.
    """

    def __init__(
        self, initial_page, page_processors=None, *, scraper: scrapelib.Scraper = None
    ):
        self.initial_page = initial_page
        if page_processors is None:
            self.page_processors = []
        elif not isinstance(page_processors, (list, tuple)):
            self.page_processors = [page_processors]
        else:
            self.page_processors = page_processors
        if not scraper:
            self.scraper = scrapelib.Scraper()

    def get_new_filename(self, obj):
        return str(uuid.uuid4())

    def save_object(self, obj, output_dir):
        filename = os.path.join(output_dir, self.get_new_filename(obj))
        with open(filename, "w") as f:
            json.dump(obj, f)

    def execute(self, output_dir: str = None) -> None:
        count = 0
        if not output_dir:
            dirn = 1
            today = datetime.date.today().strftime("%Y-%m-%d")
            while True:
                try:
                    output_dir = f"_scrapes/{today}/{dirn:03d}"
                    os.makedirs(output_dir)
                    break
                except FileExistsError:
                    dirn += 1
        else:
            try:
                os.makedirs(output_dir)
            except FileExistsError:
                if len(glob.glob(output_dir + "/*")):
                    raise FileExistsError(f"{output_dir} exists and is not empty")

        count = 0
        for data in self.yield_items():
            if isinstance(data, dict):
                dd = data
            else:
                dd = data.to_dict()
            self.save_object(dd, output_dir=output_dir)
        print(f"success: wrote {count} objects to {output_dir}")

    def yield_items(self):
        self.initial_page._fetch_data(self.scraper)
        for i, item in enumerate(self.initial_page.process_page()):
            data = item
            try:
                for pp_cls in self.page_processors:
                    page = pp_cls(data)
                    page._fetch_data(self.scraper)
                    data = page.process_page()
            except HandledError:
                continue
            yield data


class Source:
    pass


class URL(Source):
    """
    Defines a resource to fetch via URL, particularly useful for handling non-GET requests.

    :param url: URL to fetch
    :param method: HTTP method to use, defaults to "GET"
    :param data: POST data to include in request body.
    :param headers: dictionary of HTTP headers to set for the request.
    """

    def __init__(
        self, url: str, method: str = "GET", data: dict = None, headers: dict = None
    ):
        self.url = url
        self.method = method
        self.data = data
        self.headers = headers

    def get_response(
        self, scraper: scrapelib.Scraper
    ) -> Optional[requests.models.Response]:
        return scraper.request(
            method=self.method, url=self.url, data=self.data, headers=self.headers
        )

    def __str__(self) -> str:
        return self.url


class NullSource(Source):
    """
    Special class to set as a page's :py:attr:`source` to indicate no HTTP request needs
    to be performed.
    """

    def get_response(
        self, scraper: scrapelib.Scraper
    ) -> Optional[requests.models.Response]:
        return None

    def __str__(self) -> str:
        return self.__class__.__name__



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['spatula']

package_data = \
{'': ['*']}

install_requires = \
['Sphinx[docs]>=3.4.3,<4.0.0',
 'attrs[attrs]>=20.3.0,<21.0.0',
 'click>=7.1.2,<8.0.0',
 'cssselect>=1.1.0,<2.0.0',
 'ipython[shell]>=7.19.0,<8.0.0',
 'lxml>=4.6.2,<5.0.0',
 'scrapelib>=1.2.0,<2.0.0',
 'sphinx-click[docs]>=2.5.0,<3.0.0',
 'sphinx-rtd-theme[docs]>=0.5.1,<0.6.0']

entry_points = \
{'console_scripts': ['spatula = spatula.cli:cli']}

setup_kwargs = {
    'name': 'spatula',
    'version': '0.4.0',
    'description': 'lightweight structured scraping toolkit',
    'long_description': '# spatula\n\n**spatula** is a lightweight scraping framework, designed to encourage best practices and ease development and testing of scrapers.\n\n[![PyPI version](https://badge.fury.io/py/spatula.svg)](https://badge.fury.io/py/spatula)\n\n[![](https://readthedocs.org/projects/spatula/badge/?version=latest&style=flat)](https://spatula.readthedocs.org)\n',
    'author': 'James Turk',
    'author_email': 'dev@jamesturk.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jamesturk/spatula/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
